import os
import time
import numpy as np
from collections import deque
from torchPIV.PlotterFunctions import PIVparams, natural_keys
from PyQt5.QtCore import QObject, pyqtSignal, QThread, QTimer
from torchPIV.PlotterFunctions import save_table
from torchPIV.PIVbackend import OfflinePIV, free_cuda_memory



class Worker(QObject):
    finished = pyqtSignal(dict)
    progress = pyqtSignal(int)
    output   = pyqtSignal(dict)
    failed   = pyqtSignal()

class PIVWorker(Worker):
    def __init__(self, piv_params: PIVparams, *args, **kwargs) -> None:
        super().__init__(*args, parent=None, **kwargs)
        self.folder     = piv_params.folder
        self.piv_params = piv_params
        self.avg_u      = None
        self.avg_v      = None
        self.is_paused  = False
        self.is_running = True
        self.idx        = 0

    def run(self):
        """Long PIV task.
        Emits: 
        x_grid, y_grid, U, V: np.ndarray
        """
        piv_gen = OfflinePIV(
            folder=self.folder,
            device=self.piv_params.device,
            file_fmt=self.piv_params.file_fmt,
            wind_size=self.piv_params.wind_size,
            overlap=self.piv_params.overlap,
            iter_mod=self.piv_params.iter_mod,
            multipass=self.piv_params.iterations,
            dt=self.piv_params.dt,
            scale = self.piv_params.scale,
            iter_scale=self.piv_params.iter_scale
        )
        if len(piv_gen) == 0:
            self.failed.emit()
            return
        u_inst = []
        v_inst = [] 
        x = y = u = v = np.zeros((10, 10))
        start = time.time()
        for i, out in enumerate(piv_gen()):
            while self.is_paused and self.is_running: 
                time.sleep(0)
            if not self.is_running:
                break

            x, y, u, v = out

            u_inst.append(u.astype(np.float64))
            v_inst.append(v.astype(np.float64))
            self.progress.emit((i + 1)/len(piv_gen)*100)
            output = {
            "x[mm]": x,
            "y[mm]": y,
            "Vx[m/s]": u,
            "Vy[m/s]": v
            }
            if self.piv_params.save_opt == "Save all":
                name = os.path.basename(os.path.normpath(self.folder))
                save_table(f"{name}_pair.txt", self.piv_params.save_dir, output.copy())
            self.output.emit(output)
        
        if self.avg_u is None:
            self.avg_u = np.zeros_like(u, dtype=np.float64)
            self.avg_v = np.zeros_like(v, dtype=np.float64)

        print(f"Avg PIV time {((time.time() - start)/(i+1)*1000):.0f} ms")
        self.progress.emit(0)
        if u_inst:
            u_inst = np.stack(u_inst, axis=0)
            v_inst = np.stack(v_inst, axis=0)
        self.avg_u = np.mean(u_inst, axis=0, dtype=np.float64) 
        self.avg_v = np.mean(v_inst, axis=0, dtype=np.float64)
        self.progress.emit(25)
        uu = np.mean((u_inst - self.avg_u)**2, axis=0, dtype=np.float64)
        self.progress.emit(50)
        vv = np.mean((v_inst - self.avg_v)**2, axis=0, dtype=np.float64)
        self.progress.emit(75)
        uv = np.mean((u_inst - self.avg_u)*(v_inst - self.avg_v), axis=0, dtype=np.float64)
        self.progress.emit(100)
        out = (x, y, self.avg_u, self.avg_v)


        mid_i, mid_j = x.shape[-2]//2, x.shape[-1]//2
        dx = (x[mid_i, mid_j + 1] - x[mid_i, mid_j]) / 1000
        dy = (y[mid_i + 1, mid_j] - y[mid_i, mid_j]) / 1000
        dUy, dUx = np.gradient(self.avg_u, dx, dy, edge_order=2) # U  - Xcomp
        dVy, dVx = np.gradient(self.avg_v, dx, dy, edge_order=2) # V  - Ycomp
        table = {
            "x[mm]": x,
            "y[mm]": y,
            "Vx[m/s]": self.avg_u,
            "Vy[m/s]": self.avg_v,
            "(vx-Vx)(vy-Vy)[m^2/s^2]": uv,
            "(vx-Vx)^2[m^2/s^2]": uu,
            "(vy-Vy)^2[m^2/s^2]": vv,
            "dVx/dx[1/s]": dUx,
            "dVx/dy[1/s]": dUy,
            "dVy/dx[1/s]": dVx,
            "dVy/dy[1/s]": dVy,
            "W[1/s]": (dVx - dUy),
            "S[1/s]": (dVx + dUy),
        }
        free_cuda_memory()
        name = os.path.basename(os.path.normpath(self.folder))
        if self.piv_params.save_opt != "Dont save":
            save_table(f"{name}_statistics.txt", self.piv_params.save_dir, table.copy())
        self.finished.emit(table)



class OnlineWorker(Worker):

    def __init__(self, folder, piv_params: PIVparams, *args, **kwargs) -> None:
        super().__init__(*args, parent=None, **kwargs)
        self.folder = folder
        self.piv_params = piv_params
        self.avg_u      = None
        self.avg_v      = None
        self.is_paused  = False
        self.is_running = True
        self._pair    = []
        self.timer = QTimer()
        self.watchman = WatchMan(piv_params.folder, piv_params.file_fmt)

    def run(self):
        """Long PIV task.
        Emits: 
        img, x_grid, y_grid, U, V: np.ndarray
        """
        while self.is_running:
            while self.is_paused and self.is_running: 
                time.sleep(0)



class WatchMan(QObject):
    def __init__(self, folder: str, file_fmt: str, *args, **kwargs) -> None:
        super().__init__(*args, parent=None, **kwargs)
        self.folder = folder
        self.file_fmt = file_fmt
        self.filenames = {os.path.join(self.folder, name) for name 
            in os.listdir(folder) if name.endswith(file_fmt)}
        self.img_pairs: list = []

    def update(self):
        filenames = {os.path.join(self.folder, name) for name 
            in os.listdir(self.folder) if name.endswith(self.file_fmt)}
        new_files = [*filenames.difference(self.filenames)]
        self.filenames = filenames
        self.set_image_pairs(new_files=new_files)

    def set_image_pairs(self, new_files: list):
        new_files.sort(key=natural_keys)
        if len(new_files) % 2 == 0 and new_files[0].endswith("_a" + self.file_fmt):
            self.img_pairs = list(zip(new_files[::2], new_files[1::2]))
        elif len(new_files) % 2 != 0 and new_files[0].endswith("_a" + self.file_fmt):
            self.img_pairs = list(zip(new_files[:-1:2], new_files[1:-1:2]))
        elif len(new_files) % 2 != 0 and new_files[0].endswith("_b" + self.file_fmt):
            self.img_pairs = list(zip(new_files[1::2], new_files[2::2]))
        elif len(new_files) % 2 == 0 and new_files[0].endswith("_b" + self.file_fmt):
            self.img_pairs = list(zip(new_files[1:-1:2], new_files[2:-1:2]))
    def get_image_pairs(self):
        if self.img_pairs:
            return self.img_pairs
        


import os
import re
import pandas as pd
import numpy as np
import json
from PyQt5.QtWidgets import QMessageBox

def show_message(message: str) -> None:
    msgbox = QMessageBox()
    msgbox.setIcon(QMessageBox.Information)
    msgbox.setText(message)
    msgbox.setStandardButtons(QMessageBox.Ok)
    msgbox.buttonClicked.connect(msgbox.close)
    msgbox.exec_()

def uniquify(path: str):
    filename, extension = os.path.splitext(path)
    counter = 1

    while os.path.exists(path):
        path = filename + " (" + str(counter) + ")" + extension
        counter += 1

    return path


def atoi(text):
    return int(text) if text.isdigit() else text


def natural_keys(text):
    '''
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    '''
    return [ atoi(c) for c in re.split(r'(\d+)', text) ]

def set_width(obj: object, target_type: type, width: int):
    '''
    Helper function to set size of object's inner widgets
    '''
    if "setFixedWidth" in dir(target_type):
        for key, val in obj.__dict__.items():
            if isinstance(val, target_type):
                obj.__dict__[key].setFixedWidth(width)


def save_table(name, path, data: dict, sep:str=', '):
    for key in data.keys():
        data[key] = data[key].reshape(-1)
    data = pd.DataFrame(data)
    
    if not os.path.exists(path):
        os.mkdir(path)
    path = uniquify(os.path.join(path, name))
    np.savetxt(path, data.values, 
        delimiter=sep, header=sep.join(data.columns), 
        comments='', fmt="%.6f")


def make_name(name: str, key: str, orientation: bool) -> tuple:
    orientation = "Hor" if orientation else "Vert"
    name = os.path.basename(os.path.normpath(name))
    key = key[:key.find("[")].replace("/", "_")
    filename =  f"{name}_{key}_{orientation}_profile.txt".replace(' ', '')
    curr_dir = os.getcwd()
    save_dir = os.path.join(curr_dir, "Out")
    return filename, save_dir

def autoscale_y(ax, margin=0.2):

    def get_bottom_top(line):
        xd = line.get_xdata()
        yd = line.get_ydata()
        lo,hi = ax.get_xlim()
        y_displayed = yd[((xd>lo) & (xd<hi))]
        h = np.max(y_displayed) - np.min(y_displayed)
        bot = np.min(y_displayed)-margin*h
        top = np.max(y_displayed)+margin*h

        return bot,top

    lines = ax.get_lines()
    bot,top = np.inf, -np.inf

    for line in lines:
        new_bot, new_top = get_bottom_top(line)
        if new_bot < bot: bot = new_bot
        if new_top > top: top = new_top

    ax.set_ylim(bot,top)

def find_grid(data: pd.DataFrame) -> int:
    first_key = data.keys()[0]
    values = data[first_key].values
    zero_val = values[0]
    for idx, val in enumerate(values):
        if val == zero_val and idx > 0:
            break
    return idx

def reshape_data(data: pd.DataFrame, grid: int) -> dict:
    data = {key: val.values.reshape(-1, grid) for key, val in data.items()}
    return data

class PIVparams(object):
    """
    Naive Singletone inplementation 
    Shared object

    """
    wind_size: int = 0
    overlap: int = 0 
    scale: float = 0.0
    dt: float = 0.0
    device: str = ""
    iterations: int = 1
    file_fmt: str = ""
    save_opt: str = "" 
    save_dir: str = "" 
    iter_scale: float = 2.0
    folder: str = ""
    regime: str = ""
    iter_mod:str=""

    @classmethod
    def __setattr__(cls, name, val):
        setattr(cls, name, val)

    @classmethod
    def __getattr__(cls, name):
        return getattr(cls, name)

    @classmethod
    def from_json(cls):
        """
        read json settings file
        convert to dict
        set attributes
        """
        path=os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'settings.json')
        with open(path, 'r') as file:
            data = json.load(file)
            for key, val in data.items():
                if key not in dir(cls):
                    continue
                setattr(cls, key, val)
    @classmethod
    def to_json(cls):
        """
        Read attributes
        convert to dict
        save to json file
        """
        data = {
            name: getattr(cls, name) for name in dir(cls) 
            if not callable(getattr(cls, name)) and not name.startswith("__")
        }
        path=os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'settings.json')
        with open(path, 'w') as file:
            json.dump(data, file)

class Singleton(object):
    """Alex Martelli implementation of Singleton (Borg)
    http://python-3-patterns-idioms-test.readthedocs.io/en/latest/Singleton.html"""
    _shared_state = {}

    def __init__(self):
        self.__dict__ = self._shared_state


class Database(Singleton):
    def __init__(self):
        Singleton.__init__(self)
        self._data: dict = {}

    def get(self):
        return self._data

    def set(self, data: dict):
        self._data = data
    def load(self, name):
        data = pd.read_csv(name, sep=None, engine="python")
        grid = find_grid(data)
        self._data = reshape_data(data, grid)
        _, name = os.path.split(name)
        self.name, _ = os.path.splitext(name)



import sys
import time
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

class EventPrinter(FileSystemEventHandler):
    def on_created(self, event):
        print(event.src_path, flush=True)
        pass

if __name__ == "__main__":
    path = sys.argv[1] if len(sys.argv) > 1 else '.'
    event_handler = EventPrinter()
    observer = Observer()
    observer.schedule(event_handler, path, recursive=False)
    observer.start()
    try:
        while True:
            time.sleep(1)
    finally:
        observer.stop()
        observer.join()